
/*
 * gtvg_favourite.c
 * Copyright (C) Niklas Knutsson 2007 <nq@altern.org>
 * 
 * gtvg_favourite.c is free software.
 * 
 * You may redistribute it and/or modify it under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * callbacks.h is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with gtvg_favourite.h.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "gtvg_favourite.h"

#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#ifndef __USE_XOPEN
#define __USE_XOPEN
#include <time.h>
#undef __USE_XOPEN
#else
#include <time.h>
#endif

void free_recording_occasion(struct recording_occasion *recocc)
{
	if(recocc->channel != NULL)
		g_free(recocc->channel);
	g_slice_free(struct recording_occasion, recocc);
}

void free_recording_occasion_list_item(gpointer data, gpointer user_data)
{
	struct recording_occasion *recocc = (struct recording_occasion *) data;

	free_recording_occasion(recocc);
}

gint recocc_compare_func(gconstpointer a, gconstpointer b)
{
	struct recording_occasion *recocc1 = (struct recording_occasion *) a;
	struct recording_occasion *recocc2 = (struct recording_occasion *) b;

	if(recocc1->time > recocc2->time)
		return 1;
	if(recocc1->time < recocc2->time)
		return -1;
	return 0;
}

struct GtvgFavourite *gtvg_favourite_new(void)
{
	struct GtvgFavourite *fav = g_slice_new(struct GtvgFavourite);

	gtvg_favourite_init(fav);
	return fav;
}
void gtvg_favourite_free(struct GtvgFavourite *fav)
{
	gtvg_favourite_clear(fav);
	g_slice_free(struct GtvgFavourite, fav);
}
void gtvg_favourite_init(struct GtvgFavourite *fav)
{
	fav->title = NULL;
	fav->partial_title = FALSE;
	fav->channel = NULL;
	fav->timeofday = -1;
	fav->dayofweek = -1;
	fav->time = 0;
	fav->reminder = -1;
	fav->record = -1;
	fav->prioritise = FALSE;
	fav->disabled_recordings = NULL;
	fav->prioritised_recordings = NULL;
}
void gtvg_favourite_clear(struct GtvgFavourite *fav)
{
	if(fav->title != NULL)
		g_free(fav->title);
	if(fav->channel != NULL)
		g_free(fav->channel);
	if(fav->disabled_recordings != NULL) {
		g_slist_foreach(fav->disabled_recordings, free_recording_occasion_list_item, NULL);
		g_slist_free(fav->disabled_recordings);
	}
	if(fav->prioritised_recordings != NULL) {
		g_slist_foreach(fav->prioritised_recordings, free_recording_occasion_list_item, NULL);
		g_slist_free(fav->prioritised_recordings);
	}
	gtvg_favourite_init(fav);
}
void gtvg_favourite_prioritise_recording(struct GtvgFavourite *fav, gchar * channel, time_t time)
{
	struct recording_occasion *recocc;

	if(gtvg_favourite_recording_prioritised(fav, channel, time))
		return;
	recocc = g_slice_new(struct recording_occasion);
	if(fav->channel == NULL)
		recocc->channel = g_strdup(channel);
	else
		recocc->channel = NULL;
	recocc->time = time;
	fav->prioritised_recordings = g_slist_insert_sorted(fav->prioritised_recordings, recocc, recocc_compare_func);
}
gboolean gtvg_favourite_recording_prioritised(struct GtvgFavourite *fav, gchar * channel, time_t time)
{
	GSList *list = fav->prioritised_recordings;

	while(list != NULL) {
		struct recording_occasion *recocc = (struct recording_occasion *) list->data;

		if(recocc->time == time && (recocc->channel == NULL || channel == NULL || strcmp(channel, recocc->channel) == 0)) {
			return TRUE;
		} else if(recocc->time > time) {
			break;
		}
		list = g_slist_next(list);
	}
	return FALSE;
}

gint gtvg_favourite_recording_priority(struct GtvgFavourite * fav, gchar * channel, time_t time)
{
	GSList *list = fav->prioritised_recordings;

	while(list != NULL) {
		struct recording_occasion *recocc = (struct recording_occasion *) list->data;

		if(recocc->time == time && (recocc->channel == NULL || channel == NULL || strcmp(channel, recocc->channel) == 0)) {
			return 2;
		} else if(recocc->time > time) {
			break;
		}
		list = g_slist_next(list);
	}
	if(fav->prioritise)
		return 1;
	return 0;
}
void gtvg_favourite_unprioritise_recording(struct GtvgFavourite *fav, gchar * channel, time_t time)
{
	GSList *list = fav->prioritised_recordings;

	while(list != NULL) {
		struct recording_occasion *recocc = (struct recording_occasion *) list->data;

		if(recocc->time == time && (recocc->channel == NULL || channel == NULL || strcmp(channel, recocc->channel) == 0)) {
			fav->prioritised_recordings = g_slist_delete_link(fav->prioritised_recordings, list);
			free_recording_occasion(recocc);
			break;
		} else if(recocc->time > time) {
			break;
		}
		list = g_slist_next(list);
	}
}
void gtvg_favourite_disable_recording(struct GtvgFavourite *fav, gchar * channel, time_t time)
{
	struct recording_occasion *recocc;

	if(gtvg_favourite_recording_disabled(fav, channel, time))
		return;
	recocc = g_slice_new(struct recording_occasion);
	if(fav->channel == NULL)
		recocc->channel = g_strdup(channel);
	else
		recocc->channel = NULL;
	recocc->time = time;
	fav->disabled_recordings = g_slist_insert_sorted(fav->disabled_recordings, recocc, recocc_compare_func);
	gtvg_favourite_unprioritise_recording(fav, channel, time);
}
gboolean gtvg_favourite_recording_disabled(struct GtvgFavourite *fav, gchar * channel, time_t time)
{
	GSList *list = fav->disabled_recordings;

	while(list != NULL) {
		struct recording_occasion *recocc = (struct recording_occasion *) list->data;

		if(recocc->time == time && (recocc->channel == NULL || channel == NULL || strcmp(channel, recocc->channel) == 0)) {
			return TRUE;
		} else if(recocc->time > time) {
			break;
		}
		list = g_slist_next(list);
	}
	return FALSE;
}
void gtvg_favourite_enable_recording(struct GtvgFavourite *fav, gchar * channel, time_t time)
{
	GSList *list = fav->disabled_recordings;

	while(list != NULL) {
		struct recording_occasion *recocc = (struct recording_occasion *) list->data;

		if(recocc->time == time && (recocc->channel == NULL || channel == NULL || strcmp(channel, recocc->channel) == 0)) {
			fav->disabled_recordings = g_slist_delete_link(fav->disabled_recordings, list);
			free_recording_occasion(recocc);
			break;
		} else if(recocc->time > time) {
			break;
		}
		list = g_slist_next(list);
	}
}
